/******************************************************************************
* getopt_auto.c - $Revision: 1.3 $ - $Date: 2004/11/14 19:28:28 $
*
*   Copyright (C) 2004 Craig B. Agricola (craig@theagricolas.org)
*
* getopt_auto is a wrapper for the standard getopt_long() facility, just to
* make things that much easier.  It takes a data structure that describes the
* options and it takes care of the parsing of the values and printing of the
* help page. Of course, with more things being taken care of, it adds
* additional limits, but for most of my quick n' dirty projects, this is all I
* need.
* 
* BUILDING
*
* Since it is so small, I don't bother making it into a shared library; I just
* include getopt_auto.h and list getopt_auto.c as a source file when building
* things that require it.  If you want to work up the autoconf setup to get
* it to build as a shared library, feel free to do so and send it to me.
*
* USAGE
*
* See the main function conditionally compiled on the define TEST_MAIN
* below for a pretty good example.
*
* LICENSE
*
* getopt_auto is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2, or (at your option) any later version.
*
* getopt_auto is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should be able to download a copy of the GNU General Public License from
* http://www.gnu.org/licenses/gpl.txt
* If not, write to the Free Software Foundation,
* 59 Temple Place, Suite 330, Boston, MA 02111 USA.
*
******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "getopt_auto.h"

const char *arg_type_str[GETOPT_AUTO_NUM_ARG_TYPES] = { "str", "int", "dec" };

void print_wrapped_string(char *str, int column, char *spaces) {
  int c, avail_width;
  char *dup, *start, *scan, *scan_back;
  char save;
  dup = strdup(str);
  avail_width = 80 - column;
  spaces[column] = 0;
  for(start=dup; *start; start=scan) {
    for(scan=start, c=0; *scan && (c < avail_width); scan++, c++);
    if (*scan) {
      for(scan_back=scan; (*scan_back!=' ') && (scan_back!=start); scan_back--);
      if (scan_back!=start) scan=scan_back;
    }
    save = *scan;
    *scan = 0;
    printf(start);
    *scan = save;
    if (*scan) {
      printf("\n%s", spaces);
      scan++;
    }
  }
  spaces[column] = SPACE_CHAR;
  free(dup);
}

void print_option(struct getopt_auto_option_s *opt, char *spaces,
                  int have_short, int max_name_width)
{
  int data_width, help_col;
  printf("  ");
  if (have_short) {
    if (opt->short_char) printf("-%c, ", opt->short_char);
    else                 printf("    ");
  }
  printf("--%s", opt->name);
  data_width = strlen(opt->name);
  switch (opt->has_arg) {
    case 2:
      printf("[=<%s>]", arg_type_str[opt->arg_type]);
      data_width += 8;
      break;
    case 1:
      printf("=<%s>",   arg_type_str[opt->arg_type]);
      data_width += 6;
      break;
    default: break;
  }
  spaces[max_name_width - data_width + 2] = 0;
  printf("%s", spaces);
  spaces[max_name_width - data_width + 2] = SPACE_CHAR;

  help_col = max_name_width + (have_short?4:0) + 8;
  print_wrapped_string(opt->help_string, help_col, spaces);
  printf("\n");
}

void print_help(char *usage, struct getopt_auto_option_s *options) {
  struct getopt_auto_option_s *opts;
  int i, have_short = 0, max_name_width = 0;
  int explicit_help = 0, dash_h_used = 0;

  char spaces[80];
  for(i=0; i<79; i++) spaces[i]=SPACE_CHAR;
  spaces[79] = 0;

  for(opts=options; opts->name; opts++) {
    int name_width = strlen(opts->name);
    switch (opts->has_arg) {
      case 2:  name_width += 8; break;
      case 1:  name_width += 6; break;
      default:                  break;
    }
    if (name_width > max_name_width) max_name_width = name_width;
    if (!strncmp(opts->name, "help", 5)) explicit_help = 1;
    if (opts->short_char) have_short = 1;
    if (opts->short_char == 'h') dash_h_used = 1;
  }
  if ((!explicit_help) && (max_name_width < 4)) max_name_width = 4;

  printf("\nUsage:\n\n  ");
  print_wrapped_string(usage, 2, spaces);
  printf("\n\nOptions:\n\n");

  for(opts=options; opts->name; opts++)
    print_option(opts, spaces, have_short, max_name_width);
  if (!explicit_help) {
    struct getopt_auto_option_s help_opt;
    help_opt.has_arg = 0;
    help_opt.name = "help";
    help_opt.help_string = "This help page";
    if (dash_h_used) help_opt.short_char = 0;
    else             help_opt.short_char = 'h';
    print_option(&help_opt, spaces, have_short, max_name_width);
  }
  printf("\n");
}

int getopt_auto(int argc, char *argv[],
                char *usage_str,
                struct getopt_auto_option_s *auto_opts)
{
  int retval = 0;
  int i, j;
  int num_options, opt_wo_arg, long_opt_alloc, optstring_len, optstring_pos;
  int help_option_index, dash_h_used;
  char *optstring;
  struct option *long_options;
  struct getopt_auto_option_s *aop;

  num_options = 0;
  opt_wo_arg = 0;
  help_option_index = -1;
  dash_h_used = 0;
  for(aop=auto_opts; aop->name; aop++) {
    if (!strncmp(aop->name, "help", 5)) help_option_index = num_options;
    if (aop->short_char == 'h') dash_h_used = 1;
    if (aop->has_arg==0) opt_wo_arg++;
    if (!aop->storage) {
      fprintf(stderr, "%s: storage for option `--%s' not given\n",
                      argv[0], aop->name);
      retval = -2;
    }
    num_options++;
  }
  long_opt_alloc = num_options + opt_wo_arg + 1;
  if (help_option_index == -1) long_opt_alloc++;

  long_options = (struct option*) malloc (sizeof(struct option)*long_opt_alloc);
  if (!long_options) retval = -3;

  if (retval) return(retval);

  optstring_len = 0;
  for(i=0; i<num_options; i++) {
    long_options[i].name    = auto_opts[i].name;
    long_options[i].has_arg = auto_opts[i].has_arg;
    long_options[i].val     = (1 << 8) + i;
    long_options[i].flag    = 0;
    if (auto_opts[i].short_char)
      optstring_len += (1 + auto_opts[i].has_arg);
  }
  for(j=0; j<num_options; j++) {
    if (auto_opts[j].has_arg == 0) {
      char *name = (char*) malloc (strlen(auto_opts[j].name) + 4);
      if (name) sprintf(name, "no-%s", auto_opts[j].name);
      else      retval = -3;
      long_options[i].name    = name;
      long_options[i].val     = (2 << 8) + j;
      long_options[i].has_arg = 0;
      long_options[i].flag    = 0;
      i++;
    }
  }
  if (help_option_index == -1) {
    help_option_index = i++;
    long_options[help_option_index].name    = "help";
    long_options[help_option_index].has_arg = 0;
    long_options[help_option_index].val     = (1 << 8) + help_option_index;
    long_options[help_option_index].flag    = 0;
    if (!dash_h_used) optstring_len++;
  }
  long_options[i].name    = 0;
  long_options[i].has_arg = 0;
  long_options[i].val     = 0;
  long_options[i].flag    = 0;

  optstring = (char*) malloc (optstring_len+1);
  if (optstring) {
    optstring_pos = 0;
    for(i=0; i<num_options; i++) {
      if (auto_opts[i].short_char) {
        optstring[optstring_pos++] = auto_opts[i].short_char;
        if (auto_opts[i].has_arg >= 1) optstring[optstring_pos++] = ':';
        if (auto_opts[i].has_arg >= 2) optstring[optstring_pos++] = ':';
      }
    }
    if ((help_option_index >= num_options) && (!dash_h_used))
      optstring[optstring_pos++] = 'h';
    optstring[optstring_pos] = 0;
  } else
    retval = -3;

  int opt;
  while (((opt=getopt_long(argc, argv, optstring, long_options, 0)) != -1) &&
         (retval == 0))
  {
    if (opt == '?') {
      retval = -1;
    } else if (opt < 256) {
      for(i=0; i<num_options; i++)
        if (auto_opts[i].short_char == opt) opt = (1 << 8) + i;
      if ((opt < 256) && (opt == 'h')) opt = (1 << 8) + help_option_index;
      if (opt < 256)
        printf("Unknown option: %c\n", opt);
    }
    if (opt >= 256) {
      int negative_arg = (((opt >> 8) & 0x3) == 2);
      opt &= 0xff;
      if (opt == help_option_index) {
        print_help(usage_str, auto_opts);
        retval = -1;
      } else {
        if (auto_opts[opt].has_arg) {
          switch(auto_opts[opt].arg_type) {
            case GETOPT_AUTO_STRING_ARG:
              *(char**)auto_opts[opt].storage = optarg;
              break;
            case GETOPT_AUTO_INTEGER_ARG:
              if (optarg)
                *(int*)auto_opts[opt].storage = atoi(optarg);
              else
                *(int*)auto_opts[opt].storage = 1;
              break;
            case GETOPT_AUTO_DOUBLE_ARG:
              if (optarg)
                *(double*)auto_opts[opt].storage = atof(optarg);
              else
                *(double*)auto_opts[opt].storage = 1;
              break;
          }
        } else
         *(int*)auto_opts[opt].storage = (!negative_arg) ? 1 : 0;
      }
    }
  }

  free(optstring);
  for(j=0; j<opt_wo_arg; j++)
    if (long_options[num_options+j].name)
      free((char*)long_options[num_options+j].name);
  free(long_options);

  return(retval);
}

#if TEST_MAIN

int main(int argc, char *argv[]) {
  struct option_values_s {
    int    a;
    int    b;
    int    c;
    char  *strArg;
    char  *strOptArg;
    int    intArg;
    int    intOptArg;
    double decArg;
    double decOptArg;
  } opts = { 0, 1, 0, "init", "init", 42, 69, 3.1415, 2.81718 };
  struct getopt_auto_option_s options[] =
  {
    { "a",      'a', &opts.a,         0, 0,                       "Option A"           },
    { "b",      'b', &opts.b,         0, 0,                       "Option B"           },
    { "c",      'c', &opts.c,         0, 0,                       "Option C"           },
    { "str",    's', &opts.strArg,    1, GETOPT_AUTO_STRING_ARG,  "String w/ Arg"      },
    { "strOpt",   0, &opts.strOptArg, 2, GETOPT_AUTO_STRING_ARG,  "String w/ Opt Arg"  },
    { "int",    'i', &opts.intArg,    1, GETOPT_AUTO_INTEGER_ARG, "Integer w/ Arg"     },
    { "intOpt",   0, &opts.intOptArg, 2, GETOPT_AUTO_INTEGER_ARG, "Integer w/ Opt Arg" },
    { "dec",    'd', &opts.decArg,    1, GETOPT_AUTO_DOUBLE_ARG,  "Double w/ Arg"      },
    { "decOpt",   0, &opts.decOptArg, 2, GETOPT_AUTO_DOUBLE_ARG,  "Double w/ Opt Arg"  },
    { 0, 0, 0, 0, 0, 0 }
  };

  char *usage = "Test application for GetOpt_Auto";

  if (!getopt_auto(argc, argv, usage, options)) {
    printf("opts.a = %d\n", opts.a);
    printf("opts.b = %d\n", opts.b);
    printf("opts.c = %d\n", opts.c);
    printf("opts.strArg = %s\n", opts.strArg?opts.strArg:"null");
    printf("opts.strOptArg = %s\n", opts.strOptArg?opts.strOptArg:"null");
    printf("opts.intArg = %d\n", opts.intArg);
    printf("opts.intOptArg = %d\n", opts.intOptArg);
    printf("opts.decArg = %f\n", opts.decArg);
    printf("opts.decOptArg = %f\n", opts.decOptArg);
  }
}

#endif /* TEST_MAIN */
